@@ -1,6 +1,6 @@
 // Process server for soft ioc
 // David H. Thompson 8/29/2003
-// Ralph Lange 03/22/2010
+// Ralph Lange <ralph.lange@gmx.de> 2007-2016
 // GNU Public License (GPLv3) applies - see www.gnu.org
 
 #include <unistd.h>
@@ -15,6 +15,7 @@ connectionItem::connectionItem(int fd, bool readonly)
     _fd = fd;
     _readonly = readonly;
     _markedForDeletion = false;
+    _log_stamp_sent = false;
 }
 
 connectionItem::~connectionItem()
