@@ -1,9 +1,12 @@
 // Process server for soft ioc
 // David H. Thompson 8/29/2003
-// Ralph Lange <ralph.lange@gmx.de> 2007-2017
+// Ralph Lange <ralph.lange@gmx.de> 2007-2019
 // Ambroz Bizjak 02/29/2016
 // Freddie Akeroyd 2016
 // Michael Davidsaver 2017
+// Hinko Kocevar 2018
+// Klemen Vodopivec 2019
+
 // GNU Public License (GPLv3) applies - see www.gnu.org
 
 #include <vector>
@@ -46,12 +49,12 @@ bool   inDebugMode;              // This enables a lot of printfs
 bool   inFgMode = false;         // This keeps child in the foreground, tty connected
 bool   logPortLocal;             // This restricts log port access to localhost
 bool   ctlPortLocal = true;      // Restrict control connections to localhost
-bool   autoRestart = true;       // Enables instant restart of exiting child
 bool   waitForManualStart = false;  // Waits for telnet cmd to manually start child
 volatile bool shutdownServer = false;   // To keep the server from shutting down
 bool   quiet = false;            // Suppress info output (server)
 bool   setCoreSize = false;      // Set core size for child
 bool   singleEndpointStyle = true;  // Compatibility style: first non-option is endpoint
+RestartMode restartMode = restart;  // Child restart mode (restart/norestart/oneshot)
 char   *procservName;            // The name of this beast (server)
 char   *childName;               // The name of that beast (child)
 char   *childExec;               // Exec to run as child
@@ -68,10 +71,10 @@ rlim_t coreSize;                 // Max core size for child
 char   *chDir;                   // Directory to change to before starting child
 char   *myDir;                   // Directory where server was started
 time_t holdoffTime = 15;         // Holdoff time between child restarts (in seconds)
+int    childExitCode = 0;        // Child's exit code
 
 pid_t  procservPid;              // PID of server (daemon if not in debug mode)
 char   *pidFile;                 // File name for server PID
-char   defaultpidFile[] = "pid.txt";  // default
 const char *timeFormat = "%c";       // Time format string
 char   defaulttimeFormat[] = "%c";    // default
 bool   stampLog = false;         // Prefix log lines with time stamp
@@ -98,8 +101,6 @@ void mLoop();
 void OnPollTimeout();
 // Daemonizes the program
 void forkAndGo();
-// Checks the command file (existence and access rights)
-int checkCommandFile(const char *command);
 void openLogFile();
 void writeInfoFile(const std::string& infofile);
 void ttySetCharNoEcho(bool save);
@@ -119,6 +120,9 @@ void writePidFile()
     int pid = getpid();
     FILE * fp;
 
+    if ( !pidFile || strlen(pidFile) == 0 )
+        return;
+
     PRINTF("Writing PID file %s\n", pidFile);
 
     fp = fopen( pidFile, "w" );
@@ -179,6 +183,7 @@ void printHelp()
            "    --logstamp [<str>]    prefix log lines with timestamp [strftime format]\n"
            " -n --name <str>          set child's name (default: arg0 of <command>)\n"
            "    --noautorestart       do not restart child on exit by default\n"
+           " -o --oneshot             after child exits, exit the server\n"
            " -p --pidfile <str>       write PID file (for server PID)\n"
            " -P --port <endpoint>     allow control connections through telnet <endpoint>\n"
            " -q --quiet               suppress informational output (server)\n"
@@ -206,6 +211,7 @@ int main(int argc,char * argv[])
     const size_t BUFLEN = 512;
     char buff[BUFLEN];
     std::string infofile;
+    bool firstRun;
 
     time(&procServStart);             // remember start time
     procservName = argv[0];
@@ -214,7 +220,6 @@ int main(int argc,char * argv[])
     timeFormat = defaulttimeFormat;
 
     pidFile = getenv( "PROCSERV_PID" );
-    if ( !pidFile || strlen(pidFile) == 0 ) pidFile = defaultpidFile;
     if ( getenv("PROCSERV_DEBUG") != NULL ) inDebugMode = true;
 
     const int ONE_CHAR_COMMANDS = 3;  // togglerestartcmd, killcmd, logoutcmd
@@ -239,6 +244,7 @@ int main(int argc,char * argv[])
             {"logstamp",       optional_argument, 0, 'S'},
             {"name",           required_argument, 0, 'n'},
             {"noautorestart",  no_argument,       0, 'N'},
+            {"oneshot",        no_argument,       0, 'o'},
             {"pidfile",        required_argument, 0, 'p'},
             {"port",           required_argument, 0, 'P'},
             {"quiet",          no_argument,       0, 'q'},
@@ -253,7 +259,7 @@ int main(int argc,char * argv[])
         /* getopt_long stores the option index here. */
         int option_index = 0;
 
-        c = getopt_long (argc, argv, "+c:de:fhi:I:k:l:L:n:p:P:qVwx:",
+        c = getopt_long (argc, argv, "+c:de:fhi:I:k:l:L:n:op:P:qVwx:",
                          long_options, &option_index);
 
         /* Detect the end of the options. */
@@ -359,7 +365,11 @@ int main(int argc,char * argv[])
             break;
 
         case 'N':                                 // No restart of child
-            autoRestart = false;
+            restartMode = norestart;
+            break;
+
+        case 'o':                                 // Exit server when child exits
+            restartMode = oneshot;
             break;
 
         case 'R':                                 // Restrict log
@@ -450,8 +460,8 @@ int main(int argc,char * argv[])
     }
 
     if (!stampFormat) {
-        if (stampFormat) {
-            char *tmp = (char*) calloc(strlen(timeFormat)+4, 1);
+        char *tmp = (char*) calloc(strlen(timeFormat)+4, 1);
+        if (tmp) {
             sprintf(tmp, "[%s] ", timeFormat);
             stampFormat = tmp;
         } else {
@@ -459,8 +469,6 @@ int main(int argc,char * argv[])
         }
     }
 
-    if (checkCommandFile(childExec)) exit(errno);
-
     struct sigaction sig;
     memset(&sig, 0, sizeof(sig));
 
@@ -539,12 +547,12 @@ int main(int argc,char * argv[])
     if (false == inFgMode && false == inDebugMode)
     {
         forkAndGo();
-        writePidFile();
     }
     else
     {
         debugFD = 1;          // Enable debug messages
     }
+    writePidFile();
 
     if (!infofile.empty()) {
         writeInfoFile(infofile);
@@ -571,7 +579,17 @@ int main(int argc,char * argv[])
                  command );
     strncat(infoMessage1, buff, INFO1LEN-strlen(infoMessage1)-1);
     snprintf(infoMessage2, INFO2LEN, "@@@ Child \"%s\" is SHUT DOWN" NL, childName);
+    if ( logFile ) {
+	if ( -1 == logFileFD )
+            snprintf(buff, BUFLEN, "@@@ Child log file: unable to open log file %s" NL,
+                     logFile );
+	else
+            snprintf(buff, BUFLEN, "@@@ Child log file: %s" NL,
+                     logFile );
+        strncat(infoMessage1, buff, INFO1LEN-strlen(infoMessage1)-1);
+    }
 
+    firstRun = true;
     // Run here until something makes it die
     while ( ! shutdownServer )
     {
@@ -630,8 +648,16 @@ int main(int argc,char * argv[])
             // This call returns NULL if the process item lives
             if (processFactoryNeedsRestart())
             {
-                npi= processFactory(childExec, childArgv);
-                if (npi) AddConnection(npi);
+                if ((restartMode == oneshot) && !firstRun) {
+                  PRINTF("Option oneshot is set... exiting\n");
+                  shutdownServer = true;
+                } else {
+                  npi= processFactory(childExec, childArgv);
+                  if (npi) AddConnection(npi);
+                  if (firstRun) {
+                  	firstRun = false;
+                  }
+                }
             }
         } else if (-1 == ready) {             // Error
             if (EINTR != errno) {
@@ -661,8 +687,10 @@ int main(int argc,char * argv[])
 
     if(!infofile.empty())
         unlink(infofile.c_str());
-    if(pidFile)
+    if(pidFile && strlen(pidFile) > 0)
         unlink(pidFile);
+
+    return childExitCode;
 }
 
 // Connection items call this to send messages to others
@@ -759,6 +787,7 @@ void OnPollTimeout()
             snprintf(buf+strlen(buf), BUFLEN-strlen(buf),
                      " Normal exit status = %d",
                      WEXITSTATUS(wstatus));
+            childExitCode = WEXITSTATUS(wstatus);
         }
 
         if (WIFSIGNALED(wstatus)) {
@@ -866,52 +895,6 @@ void forkAndGo()
 }
 
 
-// Check to see if the command is really executable:
-// Return 0 if the file is executable
-int checkCommandFile(const char * command)
-{
-    struct stat s;
-
-    PRINTF("Checking command file validity\n");
-
-    // chdir if possible (to allow relative command)
-    if (chDir && chdir(chDir)) perror(chDir);
-
-    if (stat(command,&s)) {
-        perror(command);
-        return -1;
-    }
-
-    // Back to where we came from
-    if (chdir(myDir)) perror(myDir);
-
-    if (!S_ISREG(s.st_mode)) {
-        fprintf(stderr, "%s: %s is not a regular file\n", procservName, command);
-        return -1;
-    }
-
-// On Cygwin, permission bits are meaningless
-#ifndef __CYGWIN__
-    mode_t must_perm   =         S_IXUSR        |S_IXGRP        |S_IXOTH;
-    mode_t should_perm = S_IRUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
-    if (must_perm != (s.st_mode & must_perm)) {
-        fprintf(stderr, "%s: Error - Please change permissions on %s to at least ---x--x--x\n"
-                "procServ is not able to continue without execute permission\n",
-                procservName, command);
-        return -1;
-    }
-
-    if (should_perm != (s.st_mode & should_perm)) {
-        fprintf(stderr, "%s: Warning - Please change permissions on %s to at least -r-xr-xr-x\n"
-                "procServ may not be able to continue without read and execute permissions\n",
-                procservName, command);
-        return 0;
-    }
-#endif /* __CYGWIN__ */
-
-    return 0;
-}
-
 void openLogFile()
 {
     if (-1 != logFileFD && 1 != logFileFD) {
