@@ -1,6 +1,6 @@
 // Process server for soft ioc
 // David H. Thompson 8/29/2003
-// Ralph Lange 02/27/2012
+// Ralph Lange <ralph.lange@gmx.de> 2007-2016
 // GNU Public License (GPLv3) applies - see www.gnu.org
 
 
@@ -9,6 +9,10 @@
 
 #include "procServ.h"
 
+#ifdef __CYGWIN__
+#include <windows.h>
+#endif /* __CYGWIN__ */
+
 class processClass : public connectionItem
 {
 friend connectionItem * processFactory(char *exe, char *argv[]);
@@ -29,6 +33,10 @@ protected:
     pid_t _pid;
     static processClass * _runningItem;
     static time_t _restartTime;
+    void terminateJob();
+#ifdef __CYGWIN__
+    HANDLE _hwinjob;
+#endif /* __CYGWIN__ */
 };
 
 
